#******************************************************************************
#******************************************************************************
#                 Cryptosporidium Gp60 Analyzer (Version 1.1)
#             Written by: Christine Yanta (christine.yanta@canada.ca)
#                           Date: May 2, 2017
#   Public Health Agency of Canada - National Microbiology Laboratory Guelph
#******************************************************************************
#******************************************************************************


import io
import os

from Bio import SeqIO
from Bio.Seq import Seq

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML

from Bio.Alphabet import IUPAC

import numpy as np

import copy


global TESTING
TESTING = False

#analyzingGp60 Class Description:
#   This class keeps all the information of each Cryptosporidium sample's
#   gp60 Sanger sequence and analyzes it to classify.
#   The data is obtained from reading the given ab1 file and is then aligned
#   with the reference sequences to determine family and subtype.  It then
#   counts the number of repeats within the sequence, depending on the
#   family determined.
class analyzingGp60(object):

    def __init__(self):
        self.g = []
        self.a = []
        self.t = []
        self.c = []

        self.name=""  #name of the file
        self.species=""
        self.repeats=""

        self.oldseq = []  #before trimming the ends of the sequence
        self.seq = []    #after trimming the ends of the sequence
        self.seqLength = 0

        self.peakLoc = [] #peak locations
        self.phred_qual = [] #phred quality: 10 = 90% base call accuracy, 20=99% (log scale)

        self.beginSeq = 0
        self.endSeq = 0

        self.repeatRegionStarts=0
        self.repeatRegionEnds=0

        self.forwardSeq = True
        self.file = None





    def readFiles(self, dataFile, forw, output):
        #setting whether the sequence is the forward or reverse strand
        #(used later on know whether the reverse complement is needed)
        if forw:
            self.forwardSeq = True

        else:
            self.forwardSeq = False

        self.file = output

        #**********Beginning to work with the ab1 file given:**********#

        #retrieves the sample file name (removes directory pathway)
        self.name = dataFile.split("/")[len(dataFile.split("/"))-1]

        if TESTING:
            print("\nSequence: ", self.name) #Lets user know which sequence the program is on

        #opens the ab1 file
        handle=open(dataFile,"rb")
        record=SeqIO.read(handle, "abi")

        #retrieving base amplitude data
        self.g=np.array(record.annotations['abif_raw']['DATA9'])
        self.a=np.array(record.annotations['abif_raw']['DATA10'])
        self.t=np.array(record.annotations['abif_raw']['DATA11'])
        self.c=np.array(record.annotations['abif_raw']['DATA12'])

        #obtaining primary bases, includes N's
        self.seq=list(record.annotations['abif_raw']['PBAS2'])


        self.seqLength = len(self.seq)

        self.repeatStarts = 0
        self.repeatEnds = 0

        #peak locations
        self.peakLoc=np.array(record.annotations['abif_raw']['PLOC2'])

        #phred quality of the bases
        self.phred_qual=record.letter_annotations['phred_quality']

        #Checking quality of the sequence.  If the avg. phred quality < 20 (99% base
        #   calling certainty), then the sequence cannot be analyzed.
        quality=0

        for i in range(0, len(self.phred_qual)):
            quality += self.phred_qual[i]

        quality = quality / len(self.phred_qual)
        print(quality)

        if quality < 20:
            print("LOW QUALITY")
            self.file.write("\n")
            self.file.write("\n;>Sequence: " + self.name.split(".ab1")[0])
            self.file.write("\n;>\tPoor sequence data.  Could not analyze.  Refer to sequence manually.")

            if TESTING:
                print("\tPoor sequence data. Could not analyze. Refer to sequence manually.")

            return False

        else:
            #print("TEST")
            #calling trimSeq to trim the ends of the Sanger sequence
            self.trimSeq()
            return True






    #trimSeq() trims both ends of the Sanger sequence since the sequencing is
    #   never clear (usually messy) at the beginning and the end.  Trims based
    #   on phred_quality (10 bases in a row with 99% base calling accuracy)
    def trimSeq(self):
        qual_cutOff = 25 #99% certainty
        foundBegin = False
        foundEnd = False

        #obtain non-ambiguous sequence by fixing the N's
        self.fixN()

        #finding the repeat region
        self.findRepeatRegion()


        #Starting to trim the start of the sequence
        #   Looking for where the phred_quality is greater the cutoff (99% accuracy) for
        #   10 bases in a row.
        for x in range (0, self.seqLength):
            #if the start of the repeat region, stop trimming the sequence
            #  update where the repeat region starts
            if x == self.repeatStarts:
                for i in range (self.repeatStarts, 0, -1):
                    if self.phred_qual[i] < 15:
                        self.beginSeq = i + 1
                        break

                break

            #if phred_quality is greater than 99% base accuracy and not an N
            if self.phred_qual[x] >= qual_cutOff:
                for i in range(0,10):
                    if self.phred_qual[x+i] < qual_cutOff:
                        foundBegin=False
                        break
                    else:
                        foundBegin = True

            #if it passed 10+ bases at 99%, start trim there
            if foundBegin:
                self.beginSeq = x
                break


        #Starting to trim the end of the sequence
        #  Same algorithm as the beginning, but starting from the end
        for x in range (self.seqLength-1, self.beginSeq, -1):
            #if at the end of the repeat region, stop cutting there
            if x == self.repeatEnds:
                self.endSeq = x
                break

            #if phred quality is higher than the quality cut off, must be when 10+bases
            if self.phred_qual[x] >= qual_cutOff:
                for i in range (0, -10, -1):
                    if self.phred_qual[x+i] < qual_cutOff:
                        foundEnd=False
                        break
                    else:
                        foundEnd = True

            #if found where to cut off
            if foundEnd:
                self.endSeq = x
                break

        cutSeqLength = self.endSeq - self.beginSeq

        self.repeatStarts = self.repeatStarts - self.beginSeq
        self.repeatEnds = self.repeatEnds - self.beginSeq

        tempSeq = ['']*(cutSeqLength)
        tempPeakLoc = ['']*(cutSeqLength)
        tempPhredQual = ['']*(cutSeqLength)

        #shift everything in sequence over to trim the sequence
        for t in range (self.beginSeq, self.endSeq):
            tempSeq[t-self.beginSeq] = self.seq[t]
            tempPeakLoc[t-self.beginSeq] = self.peakLoc[t]
            tempPhredQual[t-self.beginSeq] = self.phred_qual[t]

        length = self.peakLoc[self.endSeq] - self.peakLoc[self.beginSeq]

        tempA = ['']*length
        tempG = ['']*length
        tempC = ['']*length
        tempT = ['']*length

        #copy over the amplitude values
        for i in range(0, length):
            tempA[i] = copy.copy(self.a[i+self.peakLoc[self.beginSeq]])
            tempG[i] = copy.copy(self.g[i+self.peakLoc[self.beginSeq]])
            tempC[i] = copy.copy(self.c[i+self.peakLoc[self.beginSeq]])
            tempT[i] = copy.copy(self.t[i+self.peakLoc[self.beginSeq]])


        #copy all temporary variables for analysis and store for that class
        self.seq = copy.copy(tempSeq)
        self.peakLoc = copy.copy(tempPeakLoc)
        self.phred_qual = copy.copy(tempPhredQual)
        self.a = copy.copy(tempA)
        self.g = copy.copy(tempG)
        self.c = copy.copy(tempC)
        self.t = copy.copy(tempT)
        self.seqLength = cutSeqLength


        #if the reverse sequence, make the reverse complement
        if not self.forwardSeq:
            self.peakLoc = self.peakLoc[::-1] #reverse the array
            self.phred_qual = self.phred_qual[::-1]

            revCseq=''.join(self.seq)

            temp = Seq(revCseq, IUPAC.ambiguous_dna)
            rev_complement = temp.reverse_complement()

            self.seq = copy.copy(rev_complement)

            #update where the repeat region starts as the sequence is being
            #  reversed
            start = len(self.seq) - self.repeatEnds
            end = len(self.seq) - self.repeatStarts

            self.repeatStarts = start
            self.repeatEnds = end

        #if forward sequence, just make the sequence in string form instead of array
        else:
            forwSeq = ''.join(self.seq)
            temp = Seq(forwSeq, IUPAC.ambiguous_dna)
            self.seq = copy.copy(temp)






    #fixN() finds the best base when an 'N' is encountered in the sequence by finding
    #   the maximum amplitude of all four bases at that position
    #   **Not yet implemented double peaks
    def fixN(self):
        for i in range (0, self.seqLength):
            base=self.seq[i]

            #if the base is not one of the 4 DNA base codes
            if (base!='A' and base!='G' and base!='C' and base!='T'):
                index = self.peakLoc[i] #finding peak location

                #get amplitudes for all four bases and find the max
                amplitudes=[self.a[index], self.g[index], self.c[index], self.t[index]]
                maxAmp = max(amplitudes)

                if self.a[index] == maxAmp:
                    self.seq[i] = 'A'

                elif self.g[index] == maxAmp:
                    self.seq[i] = 'G'

                elif self.c[index] == maxAmp:
                    self.seq[i] = 'C'

                elif self.t[index] == maxAmp:
                    self.seq[i] = 'T'


    #*************FINDING POSSIBLE REPEAT REGION***************************
    #finding TCA region algorithm:
    #   Looks through entire sequence for the repeat region, always keeping
    #   track of where the longest TC_ (or AG_ for reverse) repeat is
    def findRepeatRegion(self):
        repeatStart = 0
        maxCount = 0

        tempCounter = 0
        inRepeat = False

        repeatForw = ['T','C']
        repeatRev = ['G', 'A']

        index=0

        #loops through entire sequence once
        while True:
            #once entire sequence has been looked at, exit loop
            if index >= self.seqLength:
                #if the last part of the sequence was a repeat region, update stats
                #  if appropriate
                if inRepeat and tempCounter > maxCount:
                    maxCount = tempCounter
                    repeatStart = index - (maxCount*3)

                break


            #if its a TC_ or AG_ part of the sequence (in a potential repeat region)
            if (self.forwardSeq and self.seq[index:index+2]==repeatForw) or \
                (not self.forwardSeq and self.seq[index:index+2]==repeatRev):

                #if found previous consecutive TC_ and AG_, keep adding onto how
                #  many and go to the next three bases (next maybe repeat)
                if inRepeat:
                    tempCounter += 1
                    index += 3

                #meaning start of a new potential repeat region
                else:
                    tempCounter = 1
                    index+=3
                    inRepeat=True

            #if it's the end of the repeat region
            elif inRepeat:
                inRepeat = False

                #if that was the longest region found, keep track of where it was
                if tempCounter > maxCount:
                    maxCount = tempCounter
                    repeatStart  = index - (maxCount*3)


                index+=1

            #if not in a repeat region
            else:
                index+=1


        self.repeatStarts = repeatStart
        self.repeatEnds = repeatStart + (maxCount*3)

        #if it's the reverse sequence, want to subtract one since the reverse sequence
        #  reads _GA and looking for the 'GA' portions (meaning you want to start at the
        #  _ part)
        if not self.forwardSeq:
            self.repeatStarts -= 1
            self.repeatEnds -= 1





    #determineRepeats() finds the number of repeats found in the gp60 sequence
    #   Looks for TCA, TCG, TCT, and R (defined below).  Will need to modify
    #   this function is different repeat regions are required to be found.
    def determineRepeats(self):
        seq = copy.copy(self.seq)

        #Seeing if good enough quality of sequence to determine repeats
        quality_cutOff=20
        goodQuality=False

        for i in range(self.repeatStarts, self.repeatEnds):
            if self.phred_qual[i] < quality_cutOff:
                for x in range(0, 3):
                    if i+x > self.repeatEnds:
                        break
                    if self.phred_qual[i+x] < quality_cutOff:
                        goodQuality = False
                    else:
                        goodQuality = True
                        break

            else:
                goodQuality=True

            if not goodQuality:
                break

        if goodQuality:
            numAcatca = 0   #C. parvum R
            numHomR = 0     #C. hominis R=AAGACGGTGGTAAGG
            numHomR2 = 0    #C. hominis R=AAACGGTGAAGG

            #after repeat region sequence, looking for R
            afterRepeat = copy.copy(self.seq[self.repeatEnds:])

            if (afterRepeat.find("ACATCA")==0):
                numAcatca = afterRepeat.count("ACATCA")

            else:
                numHomR = afterRepeat.count("AAGACGGTGGTAAGG")

                #If 2 numHomR are found, sometimes the third is different (still counted as same R)
                #   Reference:
                if numHomR == 2:
                    numHomR2 = afterRepeat.count("AAACGGTGAAGG")

                    numHomR += numHomR2

            #repeat region sequence
            seq = copy.copy(self.seq[self.repeatStarts:self.repeatEnds])

            numTCA = seq.count("TCA")
            numTCG = seq.count("TCG")
            numTCT = seq.count("TCT")

            repeat=""

            #piecing together the number of repeats and writing in standard format
            #  Format: A#G#T#R#
            if (numTCA!=0):
                repeat += "A" + str(numTCA)

            if (numTCG!=0):
                repeat += "G" + str(numTCG)

            if (numTCT!=0):
                repeat += "T" + str(numTCT)

            if (numAcatca!=0):
                repeat += "R" + str(numAcatca)

            if (numHomR != 0):
                repeat += "R" + str(numHomR)

            self.repeats = repeat
            return True

        else:
            self.file.write("\n")
            self.file.write("\n;>Sequence: " + self.name.split(".ab1")[0])
            self.file.write("\n;>\tCould not classify repeat region. Poor sequence data.  Check manually.")

            if TESTING:
                print("\tCould not classify repeat region. Poor sequence data.  Check manually.")

            self.repeats = ""
            return False


    def determineFamily(self):
        # Filename to write
        filename = "query.txt"

        # Open the file with writing permission
        myfile = open(filename, 'w')

        # Write a line to the file
        myfile.write(''.join(self.seq[self.repeatEnds:]))

        # Close the file
        myfile.close()

        blastn_cline = NcbiblastnCommandline(cmd='blastn', task='blastn',query="query.txt", dust='yes',
                                             db=os.path.dirname(__file__)+"/reference_database/gp60_ref.fa", reward=1, penalty=-2,evalue=0.00001, outfmt=5, out="result.xml")
        blastn_cline
        stdout, stderr = blastn_cline()

        result_handle = open("result.xml")
        blast_records = NCBIXML.parse(result_handle)
        blast_record = next(blast_records)


        hit = blast_record.alignments[0].title
        data = hit.split(" ")[1]

        self.species = data





    #printFasta() prints the results of the alignment and the repeat
    #   region for each sample. It then prints the fasta sequence for
    #   each sample that can succesfully be analyzed
    def printFasta(self, file, contig):

        #IF WANTING TO PRINT TO SCREEN INSTEAD
        if TESTING and contig!="":
            print(">Contig: ", end="")
            print(self.name, " | ", end="")
            print(self.species.split("(")[0], end="")
            print(" ", self.repeats, end="")
            print()
            print(contig)


        self.file.write("\n")
        self.file.write("\n;>Sequence: " + self.name.split(".ab1")[0])

        speciesName = self.species.split("(")[0]

        file.write("\n")
        file.write(">")

        if contig != "":
            file.write("Contig: ")

        file.write(self.name.split(".ab1")[0] + "|")

        file.write(speciesName.split("|")[0])
        file.write(" " + speciesName.split("|")[1])
        file.write(self.repeats)

        if speciesName.split("|")[1] == "IIc":
            file.write(speciesName.split("|")[2])


        if contig == "":
            sequence=str(self.seq)

        else:
            sequence = contig

        counter=0
        for letter in sequence:
            if counter%100 == 0:
                file.write("\n")
                file.write(letter)
            else:
                file.write(letter)
            counter+=1
        file.write("\n")

        if TESTING:
            print(sequence)



def buildContig(s1, s2):
    s1 = s1.seq
    s2 = s2.seq


    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0
    return s1[x_longest - longest: x_longest]


def gp60_main(pathlist, fPrimers, rPrimers, typeSeq, expName):
    contig = False
    onlyForwards = False
    onlyReverse = False

    if typeSeq == 'forward':
        onlyForwards = True

    elif typeSeq == 'reverse':
        onlyReverse = True

    elif typeSeq == 'contig':
        contig = True


    fPrimers=fPrimers.replace(' ','').split(',')
    rPrimers=rPrimers.replace(' ','').split(',')

    forward = analyzingGp60()
    reverse = analyzingGp60()

    #pathlist = []

    #seqs = seqs.split(',')

    #for seq in seqs:
    #    pathlist.append(os.path.abspath(seq))

    #**************************************************************
    #added to write run parameters into result files
    file = io.StringIO()
    file.write("\n;>**********************************************************************")
    file.write("\n;>SEQUENCE ANALYSIS INPUT PARAMETERS:")
    file.write("\n  ;>Reference File: gp60_ref.fa")
    file.write("\n  ;>Forward Primer: " + str(fPrimers))
    file.write("\n  ;>Reverse Primer: " + str(rPrimers))
    file.write("\n;>**********************************************************************")
    file.write("\n;>Program Results:\n")
    #**************************************************************

    pathlist.sort()
    if contig:
        if len(pathlist)%2 == 0:
            for idx, path in enumerate(pathlist):
                for i in range(0, len(fPrimers)):
                    if fPrimers[i] in path:
                        forwSeq = forward.readFiles(path, True, file)
                        break
                    else:
                        forwSeq = True

                for i in range(0, len(rPrimers)):
                    if len(pathlist) <= idx+1:
                        print("Ensure primer names entered match those in the sequencing file name.")
                        exit()

                    if rPrimers[i] in pathlist[idx+1]:
                        revSeq = reverse.readFiles(pathlist[idx+1], False, file)
                        pathlist.remove(pathlist[idx+1])
                        break
                    else:
                        revSeq = False


                if forwSeq and revSeq:
                    forward.determineFamily()
                    reverse.determineFamily()

                    #if the crypto subfamily was found, find repeat region
                    if forward.species == reverse.species and forward.species != "":
                        contig = buildContig(forward, reverse)


                        #as of now, ubiquitum does not have a repeat region
                        if "ubiquitum" not in forward.species:
                            foundRepeat = forward.determineRepeats()

                            if foundRepeat:
                                forward.printFasta(file,contig)

                            else:
                                fullname = forward.species.split("(")[0]
                                splitname = fullname.split("|")

                                if splitname[1] == "IIc":
                                    file.write("\n;>\tFamily: " + splitname[0] + "|" + splitname[1] + splitname[2])

                                else:
                                    file.write("\n;>\tFamily: " + splitname[0] + "|" + splitname[1])

                                print("\tFamily:", forward.species.split("(")[0])

                        else:
                            forward.printFasta(file, contig)

                    else:
                        if TESTING:
                            print("Forward != Reverse Species for Contig")

                        file.write("\n")
                        file.write("\n;>Sequence: " + forward.name.split(".ab1")[0])
                        file.write("\n;>Sequence: " + reverse.name.split(".ab1")[0])
                        file.write("\n;Could not build contig. Could not correctly identify species.")
                        file.write("\n;Please review the sequences manually.")
        else:
            file.write("\nCannot find all paired forward and reverse files.  Make sure all files are included to produce the contig.")

    else:
        for path in pathlist:
            if onlyForwards:
                seq = forward.readFiles(path, True, file)
            elif onlyReverse:
                seq = forward.readFiles(path, False, file)

            if seq:
                #print("IN")
                forward.determineFamily()

                if forward.species == "":
                    pass


                elif "ubiquitum" not in forward.species:
                    foundRepeat = forward.determineRepeats()

                    if foundRepeat:
                        forward.printFasta(file,"")

                    else:
                        fullname = forward.species.split("(")[0]
                        splitname = fullname.split("|")

                        if splitname[1] == "IIc":
                            file.write("\n;>\tFamily: " + splitname[0] + "|" + splitname[1] + splitname[2])

                        else:
                            file.write("\n;>\tFamily: " + splitname[0] + "|" + splitname[1])

                        print("\tFamily:", forward.species.split("(")[0])

                else:
                    forward.printFasta(file, "")

        if TESTING:
            print("\tFinished analyzing sequence.")

    experimentName = expName + "_"

    output_report_file_name = experimentName+'crypto_typer_report.fa'
    filename = os.path.join('.', output_report_file_name)
    print(filename)
    with open(filename, 'w') as resultFile:
        resultFile.write (file.getvalue ())
    print("Report written to " + os.getcwd()+"/"+output_report_file_name + ". The gp60 run completed successfully")
    return resultFile



if __name__ == "__main__":
	gp60_main()
