#******************************************************************************
#******************************************************************************
#                 Cryptosporidium Mixed Sequence Reader  (Version 1.1)
#             Written by: Christine Yanta (christine.yanta@canada.ca)
#                           Date: July 16, 2019
#   Public Health Agency of Canada - National Microbiology Laboratory Guelph
#******************************************************************************
#******************************************************************************
import io
import os

from Bio import SeqIO

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML

import numpy as np
import math


global minAmp
minAmp = 0


global TESTING
TESTING = False


############################################################################
import difflib

def get_overlap(s1, s2):
    s = difflib.SequenceMatcher(None, s1, s2)
    pos_a, pos_b, size = s.find_longest_match(0, len(s1), 0, len(s2))

    return s1[pos_a:pos_a+size]
##############################################################################
# IUPAC Code:
#	R = A or G
#	Y = C or T
#	S = G or C
#	W = A or T
#	K = G or T
#	M = A or C
#	B = C or G or T (not A)
# 	D = A or G or T (not C)
#	H = A or C or T (not G)
# 	V = A or C or G (not T)
#	N = any
##############################################################################

#iupac() determines the IUPAC base when there are multiple possibilities.  By
#   following the IUPAC code, it returns the new possible base depending on the
#   current base and the base that is wanting to be added
def addIupac(currBase, base):
    if base == 'A':
        if currBase == 'C':
            return 'M'
        elif currBase == 'G':
            return 'R'
        elif currBase == 'T':
            return 'W'
        elif currBase == 'Y':
            return 'H'
        elif currBase == 'S':
            return 'V'
        elif currBase == 'K':
            return 'D'
        elif currBase == 'B':
            return 'N'

    elif base == 'C':
        if currBase == 'G':
            return 'S'
        elif currBase =='T':
            return 'Y'
        elif currBase == 'A':
            return 'M'
        elif currBase == 'R':
            return 'V'
        elif currBase == 'W':
            return 'H'
        elif currBase == 'K':
            return 'B'
        elif currBase == 'D':
            return 'N'

    elif base == 'G':
        if currBase == 'A':
            return 'R'
        elif currBase == 'C':
            return 'S'
        elif currBase == 'T':
            return 'K'
        elif currBase == 'Y':
            return 'B'
        elif currBase == 'W':
            return 'D'
        elif currBase == 'M':
            return 'V'
        elif currBase == 'H':
            return 'N'

    elif base == 'T':
        if currBase == 'A':
            return 'W'
        elif currBase == 'C':
            return 'Y'
        elif currBase == 'G':
            return 'K'
        elif currBase == 'R':
            return 'D'
        elif currBase == 'S':
            return 'B'
        elif currBase == 'M':
            return 'H'
        elif currBase == 'V':
            return 'N'


def subIupac(currBase, base):
    if currBase == "A" or currBase == "T" or currBase == "G" or currBase == "C":
        return currBase

    elif currBase == 'R':
        if base == 'A':
            return 'G'
        elif base == 'G':
            return 'A'
        else:
            return currBase

    elif currBase == 'Y':
        if base == 'C':
            return 'T'
        elif base == 'T':
            return 'C'
        else:
            return currBase

    elif currBase == 'S':
        if base == 'G':
            return 'C'
        elif base == 'C':
            return 'G'
        else:
            return currBase

    elif currBase == 'W':
        if base == 'A':
            return 'T'
        elif base == 'T':
            return 'A'
        else:
            return currBase

    elif currBase == 'K':
        if base == 'G':
            return 'T'
        elif base == 'T':
            return 'G'
        else:
            return currBase

    elif currBase == 'M':
        if base == 'A':
            return 'C'
        elif base == 'C':
            return 'A'
        else:
            return currBase

    elif currBase == 'N':
        if base == 'A':
            return 'B'
        elif base == 'C':
            return 'D'
        elif base == 'G':
            return 'H'
        elif base == 'T':
            return 'V'

    elif currBase == 'B':
        if base == 'T':
            return 'R'
        elif base == 'G':
            return 'Y'
        elif base == 'C':
            return 'K'

    elif currBase == 'D':
        if base == 'A':
            return 'K'
        elif base == 'G':
            return 'W'
        elif base == 'T':
            return 'R'

    elif currBase == 'H':
        if base == 'A':
            return 'Y'
        elif base == 'C':
            return 'W'
        elif base == 'T':
            return 'M'

    elif currBase == 'V':
        if base == 'A':
            return 'S'
        elif base == 'C':
            return 'R'
        elif base == 'G':
            return 'M'

#MixedSeq Class Description:
#   This class keeps all the information of each Cryptosporidium sample's
#   18S Sanger sequence and analyzes it to classify.
#   The data is obtained from reading the given ab1 file and is then aligned
#   with the reference sequences to determine family and subtype.  It then
#   counts the number of repeats within the sequence, depending on the
#   family determined.
class MixedSeq(object):

    def __init__(self):
        self.g = []
        self.a = []
        self.t = []
        self.c = []

        self.name=""  #name of the file
        self.species1=""
        self.species2=""

        self.species1Seq=""
        self.species2Seq=""

        self.oldseq = []  #before trimming the ends of the sequence
        self.seq = []    #after trimming the ends of the sequence
        self.seqLength = 0

        self.peakLoc = [] #peak locations
        self.phred_qual = [] #phred quality: 10 = 90% base call accuracy, 20=99% (log scale)

        self.forwardSeq = True
        self.file = None


    def readFiles(self, dataFile, forw, output):
        #setting whether the sequence is the forward or reverse strand
        #(used later on know whether the reverse complement is needed)
        if forw:
            self.forwardSeq = True

        else:
            self.forwardSeq = False

        self.file = output

        #**********Beginning to work with the ab1 file given:**********#

        #retrieves the sample file name (removes directory pathway)
        self.name = dataFile.split("/")[len(dataFile.split("/"))-1]

        if TESTING:
            print("\nSequence: ", self.name) #Lets user know which sequence the program is on

        #opens the ab1 file
        handle=open(dataFile,"rb")
        record=SeqIO.read(handle, "abi")

        #retrieving base amplitude data
        self.g=np.array(record.annotations['abif_raw']['DATA9'])
        self.a=np.array(record.annotations['abif_raw']['DATA10'])
        self.t=np.array(record.annotations['abif_raw']['DATA11'])
        self.c=np.array(record.annotations['abif_raw']['DATA12'])

        #obtaining primary bases, includes N's

        self.seq=list(record.annotations['abif_raw']['PBAS2'].decode("UTF-8"))
        self.oldseq=list(record.annotations['abif_raw']['PBAS2'].decode("UTF-8"))

        self.seqLength = len(self.seq)


        self.repeatStarts = 0
        self.repeatEnds = 0

        #peak locations
        self.peakLoc=np.array(record.annotations['abif_raw']['PLOC2'])

        #phred quality of the bases
        self.phred_qual=record.letter_annotations['phred_quality']

        gTemp = []
        aTemp = []
        tTemp = []
        cTemp = []

        for i in range(0, len(self.peakLoc)):
            gTemp.append(self.g[self.peakLoc[i]])
            aTemp.append(self.a[self.peakLoc[i]])
            tTemp.append(self.t[self.peakLoc[i]])
            cTemp.append(self.c[self.peakLoc[i]])

        self.g=gTemp
        self.a=aTemp
        self.c=cTemp
        self.t=tTemp

    #fixN() finds the best base when an 'N' is encountered in the sequence by finding
    #   the maximum amplitude of all four bases at that position
    #   **Not yet implemented double peaks
    def fixN(self):
        qual_cutOff = 20 #99% base calling certainty
        foundBegin = False

        #starting from the beginning of the sequence upto the end
        for x in range (0, self.seqLength):
            #looking for where the phred_qual is greater than the cutoff (99%) for
            #8+ bases in a row
            if self.phred_qual[x] >= qual_cutOff and self.oldseq[x]!='N':
                for i in range(0,8):
                    if x+i > self.seqLength:
                        break
                    if self.phred_qual[x+i] < qual_cutOff and self.oldseq[x]!='N':
                        foundBegin=False
                        break
                    else:
                        foundBegin = True

            #if there was 8+ bases with 99% base calling accuracy, start the
            #  sequence there
            if foundBegin:
                self.beginSeq = x
                break

        #make new temporary arrays of the right size for trimming the beginning
        tempSeq = ['']*(self.seqLength-self.beginSeq)
        tempPeakLoc = ['']*(self.seqLength-self.beginSeq)
        tempPhredQual = ['']*(self.seqLength-self.beginSeq)

        #shift everything over to trim
        for t in range (self.beginSeq, self.seqLength):
            tempSeq[t-self.beginSeq] = self.oldseq[t]
            tempPeakLoc[t-self.beginSeq] = self.peakLoc[t]
            tempPhredQual[t-self.beginSeq] = self.phred_qual[t]

        #copy the temporary arrays to the main variables for the class
        self.oldseq = tempSeq.copy()
        self.peakLoc = tempPeakLoc.copy()
        self.phred_qual = tempPhredQual.copy()
        self.seqLength -= self.beginSeq

        for i in range (0, self.seqLength):
            base=self.oldseq[i]

            #if the base is not one of the 4 DNA base codes
            if (base!='A' and base!='G' and base!='C' and base!='T'):
                index = self.peakLoc[i] #finding peak location

                if index > len(self.a):
                    self.oldseq = self.oldseq[:i]
                    break
                #get amplitudes for all four bases and find the max
                amplitudes=[self.a[index], self.g[index], self.c[index], self.t[index]]
                maxAmp = max(amplitudes)

                if self.a[index] == maxAmp:
                    self.oldseq[i] = 'A'

                elif self.g[index] == maxAmp:
                    self.oldseq[i] = 'G'

                elif self.c[index] == maxAmp:
                    self.oldseq[i] = 'C'

                elif self.t[index] == maxAmp:
                    self.oldseq[i] = 'T'

    def findHeteroBases(self):
        maxLRI = 4
        newSeq = []

        for i in range(0, self.seqLength):
            a = self.a[i]
            g = self.g[i]
            c = self.c[i]
            t = self.t[i]

            maxAmp = max(a,g,c,t)

            if a != 0:
                max_A = math.log((maxAmp/a),2)
            else:
                max_A = 0

            if g != 0:
                max_G = math.log((maxAmp/g),2)
            else:
                max_G = 0

            if c != 0:
                max_C = math.log((maxAmp/c),2)
            else:
                max_C = 0

            if t != 0:
                max_T = math.log((maxAmp/t),2)
            else:
                max_T = 0

            LRI = max(max_A, max_G, max_C, max_T)

            if LRI <= maxLRI and LRI != 0:
                if a == maxAmp:
                    firstBase = "A"
                elif g == maxAmp:
                    firstBase = "G"
                elif c == maxAmp:
                    firstBase = "C"
                elif t == maxAmp:
                    firstBase = "T"

                if max_A == LRI:
                    secondBase = "A"
                elif max_G == LRI:
                    secondBase = "G"
                elif max_T == LRI:
                    secondBase = "T"
                elif max_C == LRI:
                    secondBase = "C"

                iupacBase = addIupac(firstBase, secondBase)
                newSeq.append(iupacBase)

            else:
                newSeq.append(self.seq[i])

        self.seq = newSeq



    def determineAllTypes(self, file):
        # Filename to write
        filename = "query.txt"

        # Open the file with writing permission
        myfile = open(filename, 'w')
        oldSeq = ''.join(self.seq)
        index = oldSeq.find("GGTG")
        seq = oldSeq[index:]

        # Write a line to the file
        myfile.write(''.join(seq))

        # Close the file
        myfile.close()


        blastn_cline = NcbiblastnCommandline(cmd='blastn', task='blastn',query="query.txt",
                                             dust='yes',db= os.path.dirname(__file__)+"/reference_database/msr_ref.fa",
                                             reward=1, penalty=-2,evalue=0.00001, outfmt=5, out="result.xml")

        stdout, stderr = blastn_cline()



        result_handle = open("result.xml")
        blast_records = NCBIXML.parse(result_handle)
        blast_record = next(blast_records)


        hit = blast_record.alignments[0].title
        data = hit.split(" ")[0]


        self.species1 = data
        referenceSeq = os.popen("blastdbcmd -db "+os.path.dirname(__file__)+"/reference_database/msr_ref.fa -entry '{0}'".format(data)).read()


        refSeq = ''.join(referenceSeq.split("\n")[1:])
        self.species1Seq = refSeq
        newseq = []

        for c in range(0, len(seq)):
            newseq.append(subIupac(seq[c], refSeq[c]))

        # Filename to write
        filename = "query.txt"

        # Open the file with writing permission
        myfile = open(filename, 'w')

        # Write a line to the file
        myfile.write(''.join(newseq))

        # Close the file
        myfile.close()


        blastn_cline = NcbiblastnCommandline(cmd='blastn', task='blastn',
                                             query="query.txt",
                                             dust='yes',
                                             db=os.path.dirname(__file__)+"/reference_database/msr_ref.fa",
                                             reward=1, penalty=-2,evalue=0.00001, outfmt=5, out="result.xml")
        blastn_cline
        stdout, stderr = blastn_cline()

        result_handle = open("result.xml")
        blast_records = NCBIXML.parse(result_handle)
        blast_record = next(blast_records)


        hit = blast_record.alignments[0].title
        data = hit.split(" ")[0]

        self.species2 = data
        self.species2Seq = newseq

        

        if TESTING:
            print("Species: " + self.species2)


        if self.species1 == self.species2:
            self.outputResults(False, file)

        else:
            self.outputResults(True, file)


    def outputResults(self, mixed, file):
        if TESTING:
            print("\n;>Sequence: " + self.name.split(".ab1")[0])

        file.write(">")
        file.write("\n;>Sequence: " + self.name.split(".ab1")[0])

        if mixed == False:
            if self.species1.split("|")[0] == "C.hominis":
                file.write("\n;>Species: " + self.species1.split("|")[0])

                counter=0
                for letter in self.species1Seq:
                    if counter%100 == 0:
                        file.write("\n")
                        file.write(letter)
                    else:
                        file.write(letter)
                    counter+=1

                file.write("\n")

                if TESTING:
                    print("\n;>Species1: " + self.species1.split("|")[0])
                    print(self.species1Seq)


            else:
                self.fixN()

                file.write("\n;>Species: " + self.species1.split("|")[0])

                counter=0
                for letter in self.oldseq:
                    if counter%100 == 0:
                        file.write("\n")
                        file.write(letter)
                    else:
                        file.write(letter)
                    counter+=1

                file.write("\n")

                if TESTING:
                    print("\n;>Species1: " + self.species1.split("|")[0])
                    print(''.join(self.oldseq))

        else:
            file.write("\n;>Species1: " + self.species1.split("|")[0])

            counter=0
            for letter in self.species1Seq:
                if counter%100 == 0:
                    file.write("\n")
                    file.write(letter)
                else:
                    file.write(letter)
                counter+=1

            file.write("\n")

            file.write("\n;>Species2: " + self.species2.split("|")[0])

            counter=0
            for letter in self.species2Seq:
                if counter%100 == 0:
                    file.write("\n")
                    file.write(letter)
                else:
                    file.write(letter)
                counter+=1

            file.write("\n")

            if TESTING:
                print("\n;>Species1: " + self.species1.split("|")[0])
                print(self.species1Seq)

                print("\n;>Species2: " + self.species2.split("|")[0])
                print(self.species2Seq)


def msr_main(pathlist, forwardP, reverseP, typeSeq, expName):
    #pathlist = pathlist.split()
    #pathlist = []

    #print(seqs)
    #seqs = seqs.split(',')

    #for seq in seqs:
    #    pathlist.append(seq)
    #print(pathlist);exit(1)

    #**************************************************************
    #added to write run parameters into result files
    file = io.StringIO()
    file.write("\n;>****************************************************************************")
    file.write("\n;>MIXED SEQUENCE ANALYSIS INPUT PARAMETERS:")
    file.write("\n  ;>Reference File: " + "msr_ref.fa")
    file.write("\n  ;>Forward Primer: " + forwardP)
    file.write("\n  ;>Reverse Primer: " + reverseP)
    file.write("\n;>****************************************************************************")
    file.write("\n;>Program Results:\n")
    #**************************************************************

    pathlist.sort()
    pathlist = [path for path in pathlist if forwardP in path or reverseP in path] #select only files matching the primers

    if len(pathlist) == 0:
        exit("No files in the input are matching the forward or reverse primer. Aborting.")


    if len(pathlist)%2 != 0 and typeSeq == 'contig':
        file.write("\nError: Need to include both forward and reverse sequences of all samples to produce contig.")
        file.write("\nSequences files given:\n")

        for i in range(len(pathlist)):
            file.write(pathlist[i])
            file.write("\n")


    elif typeSeq=='contig':
        for idx, path in enumerate(pathlist): #needs to be fixed to accomodate >2 files in the folder
            forward = MixedSeq()
            reverse = MixedSeq()
            forward.readFiles(pathlist[idx], True, file)
            reverse.readFiles(pathlist[idx+1], False, file)

            forward.findHeteroBases()
            reverse.findHeteroBases()

            forward.determineAllTypes(file)
            break

    elif typeSeq=="forward":
        for idx, path in enumerate(pathlist):
            forward=MixedSeq()
            forward.readFiles(pathlist[idx], True, file)
            forward.findHeteroBases()
            forward.determineAllTypes(file)

    elif typeSeq=="reverse":
        for idx, path in enumerate(pathlist):
            reverse=MixedSeq()
            reverse.readFiles(pathlist[idx], False, file)
            reverse.findHeteroBases()
            reverse.determineAllTypes(file)

    experimentName = expName + "_"
    output_report_file_name = experimentName + 'crypto_typer_report.fa'
    filename = os.path.join('.', output_report_file_name )

    with open(filename, 'w') as resultFile:
        resultFile.write (file.getvalue ())

    print("Report written to "+os.getcwd()+"/"+output_report_file_name + ". The 18S run completed sucessfully.")
    return resultFile


if __name__ == "__main__":
    msr_main()
