#!/usr/bin/env python

import os
import textwrap

from crypto_typer.msr import msr_main
from crypto_typer.gp60 import gp60_main
import argparse
from crypto_typer.version import  __version__

def parse_cli_arguments():
    parser = argparse.ArgumentParser(prog="crypto_typer",
                                     description='In silico type cryptosporidium from sanger reads in AB1 format\n',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog = textwrap.dedent('''Example usage using example ab1 files included in /example folder:
    crypto_typer -i example/P17705_Crypto16-2F-20170927_SSUF_G12_084.ab1 -m gp60 -t forward -f gp60F -o test   
    crypto_typer -i example/P17705_gp60-Crypt14-1F-20170927_gp60F_G07_051.ab1 -m 18S -t forward -f SSUF -o test
    crypto_typer -i example/ -m 18S -t contig -f SSUF -r SSUR -o test
    crypto_typer -i example/ -m gp60 -t contig -f gp60F -r gp60R -o test
                                                              '''))

    parser.add_argument('--verbose', action='store_true', dest='verbose',
                        help='Turn on verbose logging [False].', required=False)
    parser.add_argument('-i','--input',  nargs=1,  required=True,
                        help="Path to directory with AB1 forward and reverse files OR path to a single AB1 file")
    parser.add_argument('-m', '--marker', type=str, required=True,
                        help="Name of the marker. Currently gp60 and 18S markers are supported")
    parser.add_argument('-t', '--seqtype', type=str, required=True,
                        help="Input sequences type. Select one option out of these three:\n"
                             "contig - both F and R sequences provided\n "
                             "forward - forward only sequence provided\n"
                             "reverse - reverse only sequence provided\n")
    parser.add_argument('-f', '--forwardprimername', type=str, required=False,
                        help="Name of the forward primer to identify forward read (e.g. gp60F, SSUF)")
    parser.add_argument('-r', '--reverseprimername', type=str, required=False,
                        help="Name of the reverse primer to identify forward read (e.g. gp60R, SSUR)")
    parser.add_argument('-o', '--outputprefix', type=str, required=False,
                        help="Output name prefix for the results (e.g. test results in test_report.fa)")
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {}'.format(__version__))

    return parser.parse_args()

#in command line: sequences, marker, contig/f/r, fname, rname, expName
def main():
    #expName = ''

    args= parse_cli_arguments()
    print(args)

    #seq_dir = sys.argv[1]
    seq_dir = args.input
    marker = args.marker
    typeSeq = args.seqtype


    expName = args.outputprefix

    if typeSeq == "contig":
        fPrimer = args.forwardprimername
        rPrimer = args.reverseprimername
        #seq_dir = [seq_dir+"/"+f for f in )]

        pathlist=list()
        for file in os.listdir(seq_dir[0]):
            pathlist.append(os.path.abspath(seq_dir[0]+"/"+file))



        if marker == "18S":
            return msr_main(pathlist, fPrimer, rPrimer, typeSeq, expName)

        elif marker == "gp60":
            return gp60_main(pathlist, fPrimer, rPrimer, typeSeq, expName)

    elif typeSeq == "forward":
        fPrimer = args.forwardprimername

        if marker == "18S":
            msr_main(seq_dir, fPrimer, "", typeSeq, expName)

        elif marker == "gp60":
            gp60_main(seq_dir, fPrimer, "", typeSeq, expName)


    elif typeSeq == "reverse":
        rPrimer = args.reverseprimername

        if marker == "18S":
            msr_main(seq_dir, "", rPrimer, typeSeq, expName)

        elif marker == "gp60":
            return gp60_main(seq_dir, "", rPrimer, typeSeq, expName)






if __name__ == "__main__":
    main()

